package com.sun.corba.se.impl.dynamicany;

import java.io.Serializable;
import org.omg.CORBA.Object;
import org.omg.DynamicAny.DynAnyFactoryPackage.InconsistentTypeCode;
import org.omg.DynamicAny.DynAnyPackage.InvalidValue;
import org.omg.CORBA.portable.OutputStream;
import org.omg.DynamicAny.DynAnyPackage.TypeMismatch;
import com.sun.corba.se.impl.corba.TypeCodeImpl;
import org.omg.CORBA.TypeCode;
import com.sun.corba.se.spi.orb.ORB;
import org.omg.CORBA.Any;
import org.omg.DynamicAny.DynAny;

abstract class DynAnyConstructedImpl extends DynAnyImpl
{
    protected static final byte REPRESENTATION_NONE = 0;
    protected static final byte REPRESENTATION_TYPECODE = 1;
    protected static final byte REPRESENTATION_ANY = 2;
    protected static final byte REPRESENTATION_COMPONENTS = 4;
    protected static final byte RECURSIVE_UNDEF = -1;
    protected static final byte RECURSIVE_NO = 0;
    protected static final byte RECURSIVE_YES = 1;
    protected static final DynAny[] emptyComponents;
    DynAny[] components;
    byte representations;
    byte isRecursive;
    
    private DynAnyConstructedImpl() {
        this(null, (Any)null, false);
    }
    
    protected DynAnyConstructedImpl(ORB orb, Any any, boolean b) {
        super(orb, any, b);
        this.components = DynAnyConstructedImpl.emptyComponents;
        this.representations = 0;
        this.isRecursive = -1;
        if (this.any != null) {
            this.representations = 2;
        }
        this.index = 0;
    }
    
    protected DynAnyConstructedImpl(ORB orb, TypeCode typeCode) {
        super(orb, typeCode);
        this.components = DynAnyConstructedImpl.emptyComponents;
        this.representations = 0;
        this.isRecursive = -1;
        if (typeCode != null) {
            this.representations = 1;
        }
        this.index = -1;
    }
    
    protected boolean isRecursive() {
        if (this.isRecursive == -1) {
            TypeCode type = this.any.type();
            if ((type instanceof TypeCodeImpl)) {
                if (((TypeCodeImpl)type).is_recursive()) {
                    this.isRecursive = 1;
                }
                else {
                    this.isRecursive = 0;
                }
            }
            else {
                this.isRecursive = 0;
            }
        }
        return this.isRecursive == 1;
    }
    
    public DynAny current_component() throws TypeMismatch {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            return null;
        }
        return this.checkInitComponents() ? this.components[this.index] : null;
    }
    
    public int component_count() {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        return this.checkInitComponents() ? this.components.length : 0;
    }
    
    public boolean next() {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (!this.checkInitComponents()) {
            return false;
        }
        ++this.index;
        if (this.index >= 0 && this.index < this.components.length) {
            return true;
        }
        this.index = -1;
        return false;
    }
    
    public boolean seek(int index) {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (index < 0) {
            this.index = -1;
            return false;
        }
        if (!this.checkInitComponents()) {
            return false;
        }
        if (index < this.components.length) {
            this.index = index;
            return true;
        }
        return false;
    }
    
    public void rewind() {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        this.seek(0);
        return;
    }
    
    @Override
    protected void clearData() {
        super.clearData();
        this.components = DynAnyConstructedImpl.emptyComponents;
        this.index = -1;
        this.representations = 0;
    }
    
    @Override
    protected void writeAny(OutputStream outputStream) {
        this.checkInitAny();
        super.writeAny(outputStream);
    }
    
    protected boolean checkInitComponents() {
        if ((this.representations & 0x4) == 0x0) {
            if ((this.representations & 0x2) != 0x0) {
                if (this.initializeComponentsFromAny()) {
                    this.representations |= 0x4;
                }
                else {
                    return false;
                }
            }
            else if ((this.representations & 0x1) != 0x0) {
                if (this.initializeComponentsFromTypeCode()) {
                    this.representations |= 0x4;
                }
                else {
                    return false;
                }
            }
        }
        return true;
    }
    
    protected void checkInitAny() {
        if ((this.representations & 0x2) == 0x0) {
            if ((this.representations & 0x4) != 0x0) {
                if (this.initializeAnyFromComponents()) {
                    this.representations |= 0x2;
                }
            }
            else if ((this.representations & 0x1) != 0x0) {
                if (this.representations == 1 && this.isRecursive()) {
                    return;
                }
                if (this.initializeComponentsFromTypeCode()) {
                    this.representations |= 0x4;
                }
                if (this.initializeAnyFromComponents()) {
                    this.representations |= 0x2;
                }
            }
        }
    }
    
    protected abstract boolean initializeComponentsFromAny();
    
    protected abstract boolean initializeComponentsFromTypeCode();
    
    protected boolean initializeAnyFromComponents() {
        OutputStream create_output_stream = this.any.create_output_stream();
        for (int i = 0; i < this.components.length; ++i) {
            if ((this.components[i] instanceof DynAnyImpl)) {
                ((DynAnyImpl)this.components[i]).writeAny(create_output_stream);
            }
            else {
                this.components[i].to_any().write_value(create_output_stream);
            }
        }
        this.any.read_value(create_output_stream.create_input_stream(), this.any.type());
        return true;
    }
    
    @Override
    public void assign(DynAny dynAny) throws TypeMismatch {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        this.clearData();
        super.assign(dynAny);
        this.representations = 2;
        this.index = 0;
        return;
    }
    
    @Override
    public void from_any(Any any) throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        this.clearData();
        super.from_any(any);
        this.representations = 2;
        this.index = 0;
        return;
    }
    
    @Override
    public Any to_any() {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        this.checkInitAny();
        return DynAnyUtil.copy(this.any, this.orb);
    }
    
    @Override
    public boolean equal(DynAny dynAny) {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (dynAny == this) {
            return true;
        }
        if (!this.any.type().equal(dynAny.type())) {
            return false;
        }
        if (!this.checkInitComponents()) {
            return false;
        }
        DynAny current_component = null;
        try {
            current_component = dynAny.current_component();
            int i = 0;
            while (i < this.components.length) {
                if (!dynAny.seek(i)) {
                    return false;
                }
                if (!this.components[i].equal(dynAny.current_component())) {
                    return false;
                }
                ++i;
            }
        }
        catch (TypeMismatch typeMismatch) {}
        finally {
            DynAnyUtil.set_current_component(dynAny, current_component);
        }
        return true;
    }
    
    @Override
    public void destroy() {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.status == 0) {
            this.status = 2;
            for (int i = 0; i < this.components.length; ++i) {
                if ((this.components[i] instanceof DynAnyImpl)) {
                    ((DynAnyImpl)this.components[i]).setStatus((byte)0);
                }
                this.components[i].destroy();
            }
        }
        return;
    }
    
    @Override
    public DynAny copy() {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        this.checkInitAny();
        try {
            return DynAnyUtil.createMostDerivedDynAny(this.any, this.orb, true);
        }
        catch (InconsistentTypeCode inconsistentTypeCode) {
            return null;
        }
    }
    
    public void insert_boolean(boolean b) throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        current_component.insert_boolean(b);
        return;
    }
    
    public void insert_octet(byte b) throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        current_component.insert_octet(b);
        return;
    }
    
    public void insert_char(char c) throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        current_component.insert_char(c);
        return;
    }
    
    public void insert_short(short n) throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        current_component.insert_short(n);
        return;
    }
    
    public void insert_ushort(short n) throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        current_component.insert_ushort(n);
        return;
    }
    
    public void insert_long(int n) throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        current_component.insert_long(n);
        return;
    }
    
    public void insert_ulong(int n) throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        current_component.insert_ulong(n);
        return;
    }
    
    public void insert_float(float n) throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        current_component.insert_float(n);
        return;
    }
    
    public void insert_double(double n) throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        current_component.insert_double(n);
        return;
    }
    
    public void insert_string(String s) throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        current_component.insert_string(s);
        return;
    }
    
    public void insert_reference(Object object) throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        current_component.insert_reference(object);
        return;
    }
    
    public void insert_typecode(TypeCode typeCode) throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        current_component.insert_typecode(typeCode);
        return;
    }
    
    public void insert_longlong(long n) throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        current_component.insert_longlong(n);
        return;
    }
    
    public void insert_ulonglong(long n) throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        current_component.insert_ulonglong(n);
        return;
    }
    
    public void insert_wchar(char c) throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        current_component.insert_wchar(c);
        return;
    }
    
    public void insert_wstring(String s) throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        current_component.insert_wstring(s);
        return;
    }
    
    public void insert_any(Any any) throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        current_component.insert_any(any);
        return;
    }
    
    public void insert_dyn_any(DynAny dynAny) throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        current_component.insert_dyn_any(dynAny);
        return;
    }
    
    public void insert_val(Serializable s) throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        current_component.insert_val(s);
        return;
    }
    
    public Serializable get_val() throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        return current_component.get_val();
    }
    
    public boolean get_boolean() throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        return current_component.get_boolean();
    }
    
    public byte get_octet() throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        return current_component.get_octet();
    }
    
    public char get_char() throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        return current_component.get_char();
    }
    
    public short get_short() throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        return current_component.get_short();
    }
    
    public short get_ushort() throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        return current_component.get_ushort();
    }
    
    public int get_long() throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        return current_component.get_long();
    }
    
    public int get_ulong() throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        return current_component.get_ulong();
    }
    
    public float get_float() throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        return current_component.get_float();
    }
    
    public double get_double() throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        return current_component.get_double();
    }
    
    public String get_string() throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        return current_component.get_string();
    }
    
    public Object get_reference() throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        return current_component.get_reference();
    }
    
    public TypeCode get_typecode() throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        return current_component.get_typecode();
    }
    
    public long get_longlong() throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        return current_component.get_longlong();
    }
    
    public long get_ulonglong() throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        return current_component.get_ulonglong();
    }
    
    public char get_wchar() throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        return current_component.get_wchar();
    }
    
    public String get_wstring() throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        return current_component.get_wstring();
    }
    
    public Any get_any() throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        return current_component.get_any();
    }
    
    public DynAny get_dyn_any() throws TypeMismatch, InvalidValue {
        if (this.status == 2) {
            throw this.wrapper.dynAnyDestroyed();
        }
        if (this.index == -1) {
            throw new InvalidValue();
        }
        DynAny current_component = this.current_component();
        if (DynAnyUtil.isConstructedDynAny(current_component)) {
            throw new TypeMismatch();
        }
        return current_component.get_dyn_any();
    }
    
    static {
        emptyComponents = new DynAny[0];
    }
}
